import os.path
import glob

import matplotlib
import matplotlib.pyplot as plt
import pickle
import itertools as it
import seaborn as sns
import networkx as nx

import pandas as pd
import numpy as np
import scipy as sp
import scipy.stats as st
import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow import keras
from keras import backend as K
tfd = tfp.distributions

class OrthogonalSymmetryCayley(object):
    def __init__(self, X=None):
        self.X = X
        self.A = tf.Variable(tf.random.uniform([X.shape[1], X.shape[1]], minval=-1, maxval=1, dtype=tf.float32))
        self.optimizer = keras.optimizers.SGD(learning_rate=.1, momentum=.5)

    @staticmethod
    def gaussian_kernel(X, Y, beta=1e-5):
        nX = tf.reduce_sum(tf.square(X), axis=1)
        nY = tf.reduce_sum(tf.square(Y), axis=1)
        nX = tf.reshape(nX, [-1, 1])
        nY = tf.reshape(nY, [1, -1])
        D = nX + nY - 2*tf.matmul(X, Y, transpose_b=True)
        return D

    def mmd(self, X, Y, beta=1e-5):
        D = self.gaussian_kernel(X, Y, beta)
        return tf.reduce_sum(tf.exp(-beta*D), axis=None)

    def loss(self, X, lamb=1):
        I = tf.eye(self.X.shape[1], dtype=tf.dtypes.float32)
        A_upper = tf.linalg.band_part(self.A, 0, -1)
        A_skew = 0.5 * (A_upper - tf.transpose(A_upper))
        U = tf.matmul((I-A_skew), tf.linalg.inv(I+A_skew))
        mmd = self.mmd(X, tf.matmul(X, U, transpose_b=True))
        penalty = tf.norm(U - I, ord='fro', axis=[-2, -1])**2

        L = mmd - lamb*penalty
        self.U = U
        return L

    def train(self):
        X = self.X
        for _ in range(10):
            training_vars = [self.A]
            with tf.GradientTape() as tape:
                L = self.loss(X)
                grad = tape.gradient(L, training_vars)
                self.optimizer.apply_gradients(zip(grad, training_vars))
            scores = tf.reduce_sum(self.gaussian_kernel(X, tf.matmul(X, self.U, transpose_b=True)), axis=1)
            num_to_remove = 10
            worst_indices = tf.argsort(scores)[:num_to_remove]
            mask = tf.logical_not(tf.reduce_any(tf.equal(tf.range(tf.shape(self.X)[0])[:, tf.newaxis], worst_indices), axis=1))
            X = tf.boolean_mask(self.X, mask)
